From: Ankur Arora <ankur.a.ar...@oracle.com>

[ Upstream commit 83b28cfe796464ebbde1cf7916c126da6d572685 ]

With PREEMPT_RCU=n, cond_resched() provides urgently needed quiescent
states for read-side critical sections via rcu_all_qs().
One reason why this was needed: lacking preempt-count, the tick
handler has no way of knowing whether it is executing in a
read-side critical section or not.

With (PREEMPT_LAZY=y, PREEMPT_DYNAMIC=n), we get (PREEMPT_COUNT=y,
PREEMPT_RCU=n). In this configuration cond_resched() is a stub and
does not provide quiescent states via rcu_all_qs().
(PREEMPT_RCU=y provides this information via rcu_read_unlock() and
its nesting counter.)

So, use the availability of preempt_count() to report quiescent states
in rcu_flavor_sched_clock_irq().

Suggested-by: Paul E. McKenney <paul...@kernel.org>
Reviewed-by: Sebastian Andrzej Siewior <bige...@linutronix.de>
Signed-off-by: Ankur Arora <ankur.a.ar...@oracle.com>
Reviewed-by: Frederic Weisbecker <frede...@kernel.org>
Signed-off-by: Paul E. McKenney <paul...@kernel.org>
Signed-off-by: Boqun Feng <boqun.f...@gmail.com>
Signed-off-by: Sasha Levin <sas...@kernel.org>
---
 kernel/rcu/tree_plugin.h | 11 +++++++----
 1 file changed, 7 insertions(+), 4 deletions(-)

diff --git a/kernel/rcu/tree_plugin.h b/kernel/rcu/tree_plugin.h
index f5ba0740f9b50..c07a841971739 100644
--- a/kernel/rcu/tree_plugin.h
+++ b/kernel/rcu/tree_plugin.h
@@ -919,13 +919,16 @@ static void rcu_preempt_check_blocked_tasks(struct 
rcu_node *rnp)
  */
 static void rcu_flavor_sched_clock_irq(int user)
 {
-       if (user || rcu_is_cpu_rrupt_from_idle()) {
+       if (user || rcu_is_cpu_rrupt_from_idle() ||
+            (IS_ENABLED(CONFIG_PREEMPT_COUNT) &&
+             (preempt_count() == HARDIRQ_OFFSET))) {
 
                /*
                 * Get here if this CPU took its interrupt from user
                 * mode or from the idle loop, and if this is not a
                 * nested interrupt.  In this case, the CPU is in
                 * a quiescent state, so note it.
                 *
                 * No memory barrier is required here because rcu_qs()
                 * references only CPU-local variables that other CPUs
@@ -919,13 +919,16 @@ static void rcu_preempt_check_blocked_tasks(struct 
                /*
                 * Get here if this CPU took its interrupt from user
-                * mode or from the idle loop, and if this is not a
-                * nested interrupt.  In this case, the CPU is in
-                * a quiescent state, so note it.
+                * mode, from the idle loop without this being a nested
+                * interrupt, or while not holding the task preempt count
+                * (with PREEMPT_COUNT=y). In this case, the CPU is in a
+                * quiescent state, so note it.
                 *
                 * No memory barrier is required here because rcu_qs()
-- 
2.39.5


Reply via email to